/////////////////////////////////////////////////////////
/*
   Minimum Z80 System mit Mostek 3801 Multi I/O Chip
   (C)Christian Julius 10.2014
   christian.julius@gmx.de

   HAL der Hardware und Peripherie

*/
/////////////////////////////////////////////////////////

#include <stdio.h>
#include <stdint.h>
#include <stdlib.h>
#include <stdbool.h>
#include "mcurses.h"
#include "main.h"
#include "driver.h"

// Segment Codes der 7-Segmentanzeige, invertiert, 0 - F
const char digits[16]= {0xc0,0xf5,0xa8,0xb0,0x95,
                        0x92,0x82,0xf4,0x80,0x90,
                        0x84,0x83,0xab,0xa1,0x8a,
                        0x8e};

// RX Buffer fr die UART
struct rx_type rx;
struct tx_type tx;

// Basis Initialisierung der Hardware, MK3081 Chip
void InitHardware()
{
    // 8255: Mode 0: Port A,B,C Output
    PIO8255_CNTRL  = 0x80;      //Mode 0: Port A,B,C Output
    PIO8255_PORT_A = 0xff;
    PIO8255_PORT_B = 0;
    PIO8255_PORT_C = 0;

    // 7-Segment ....
    SEG7 = 0xff;

    // STI konfigurieren fr Uart Ausgabe
    STI_UCR   = 0x88;           // = 10001000  UART Control Register 8N1, DIV16
    STI_RSR   = 0x01;           // RX Status - Enable RX
    STI_TSR   = 0x05;           // TX Status - TX High, TX enabled
    // Timer D...
    STI_PVR   = STI_TDDR;       // Zeiger PVR auf Timer D Data
    STI_IDR   = 0x03;           // Timer D Zeitkonstante
    STI_PVR   = STI_TCDCR;      // Zeiger auf Timer C,D Control Register
    STI_IDR   = 0x89;           // Timer D DIV 4, C = Stop
    // GPIO.....
    STI_PVR   = STI_DDR;        // Zeiger PVR auf indirect 6 (DDR)
    STI_IDR   = 0xff;           // GPIO als Output
    STI_GPIP  = 0x00;           // GPIO auf Null setzen

    STI_PVR   = STI_IERA;       // Alle Interrupts zuerst aus
    STI_IDR   = 0;
    STI_IMRA  = 0;
    STI_IPRA  = 0x00;           // Lsche Pending Ints

    STI_PVR   = MODE2_VECTOR;   // Vektor auf Mode 2 Tabelle
}

/*
    Test der Hardware

    1. ROM Checksumme bilden 16 Bit Summe, CPL + 1 Anzeige
    2. RAM Schnelltest:
           Jede zweite Speicherstelle auf 55,AA testen
    3. Extension Board Anwesenheit prfen (8255 programmieren und abfragen)
    4.

*/

// Lsche einen beliebigen Puffer
void clear_buffer(char* ptr,char len)
{
   char i;
   for (i=0;i<len;i++)
    *(ptr++) = 0;

}

void SetDigit(unsigned char x)
{
    SEG7 = digits[x & 0x0f];
}

// Auslesen des AD Wandlers
uint8_t adc_read()
{
    ADC0804_READ = 1;
    delay(10);
    return (ADC0804_READ);
}

// Lscht den TX Buffer und setzt Zeiger auf 0
void init_buffer()
{
    // RX....
    rx.wr_ptr           = 0;
    rx.rd_ptr           = 0;
    rx.bytes_to_read    = 0;
    rx.f_buf_overflow   = false;
    rx.mask             = (BUF_SIZE-1) & 0xff;

    // TX.....
    tx.wr_ptr           = 0;
    tx.rd_ptr           = 0;
    tx.bytes_to_send    = 0;
    tx.f_idle           = true;
    tx.mask             = (BUF_SIZE-1) & 0xff;
}

////////////////////////////////////////////////////////
// Zeicheneingabe von der Uart
// Darf nur aufgerufen werden, wenn rx_buf_full == true
// d..h es ist mindestens 1 Zeichen da

// Liefert 0, wenn keine Taste gedrueckt wurde, sonst die Taste
char keypress()
{
    if (rx.bytes_to_read > 0)
        return (getchar());
    else
        return (0);

}

// Zentrale Eingabe Routine von der UART (Ringpuffer)
char getchar()
{
    register char ch;

    // Warte solange bis ein Zeichen im Buffer ist, auch ewig.
    while (rx.bytes_to_read == 0);
    // Zeichen auslesen
    ch = rx.buf[rx.rd_ptr];
    // Read Pointer am Array Ende zurck setzen
    rx.rd_ptr = (rx.rd_ptr +1) & rx.mask;
    // Zeichenzaehler - 1
    CLI;
    if (rx.bytes_to_read > 0)
        rx.bytes_to_read--;
    EI;

    return(ch);
}

// Zentrale Ausgabe der UART (Ringpuffer)
void putchar(char c)
{
    static char cnt1,cnt2 = 0;

    // Warte bis TX frei fr neues Zeichen ist
    //while ((STI_TSR & 0x80) == 0);
    //STI_UDR = c;

    // Warte bis Buffer mindestens 1 Byte frei hat
    while (tx.bytes_to_send >= BUF_SIZE);

    // Ist das TX Modul grad frei und Buffer leer?
    CLI;
    if (tx.f_idle) {
        tx.f_idle = false;
        STI_UDR = c;        // Int Erzeugung anstossen
        EI;
    } else
    {   // TX bertrgt gerade, dann Byte in Buffer eintragen
        tx.buf[tx.wr_ptr] = c;
        // wr+1 am Bufferende auf 0 zurcksetzen
        tx.wr_ptr = (tx.wr_ptr +1) & tx.mask;
        CLI;
        tx.bytes_to_send++;
        EI;
    }
}

/////////////////////////////////////////////////////////
// Setze oder lsche einen GPIO Port
// Input: Port Nr, HIGH oder LOW
void Set_GPIO(uint8_t bitnr, uint8_t stat)
{
    uint8_t foo;

    foo = STI_GPIP;         // Register kann nur geladen werden

    if (stat)               // Bit ein
        foo |= (1<<bitnr);
    else                    // Bit lschen
        foo &= ~(1<<bitnr);

    STI_GPIP = foo;
}

/////////////////////////////////////////////////////////
// Setze oder lsche eine LED (0..7) am Port A
// Input: Led Nr, HIGH oder LOW
void SetLed(uint8_t lednr, uint8_t stat)
{
    uint8_t foo;

    // Read...
    foo = PIO8255_PORT_A;
    // Modify...
    if (stat==1)  // LED ein = Bit LOW
        foo &= ~(1<<lednr);
    else
        foo |= (1<<lednr);
    // Write...
    PIO8255_PORT_A = foo;
}

/////////////////////////////////////////////////////////
// Zeitverzgerung
void delay(uint16_t cnt) {
    volatile uint16_t k;
    for (k=0;k<cnt;k++);
}

/////////////////////////////////////////////////////////
// Halte CPU an und lasse LED leuchten
// Ausstieg nur durch Reset
void stop_cpu() __naked
{
    __asm
        halt
    __endasm ;

    while (1);
}

////////////////////////////////////////////////////////
// Schaltet die RX Line schnell ab oder an
void rx_fast_enable(char stat)
{
    // Nur die Hardware ein/aus, nicht aber die Ints
    if (stat)
        STI_RSR   = STI_RSR  | 0x01;            // RX ein
    else
        STI_RSR   = STI_RSR  & ~0x01;           // RX aus
}
////////////////////////////////////////////////////////
// Schaltet die TX Line schnell ab oder an
void tx_fast_enable(char stat)
{
    // Nur die Hardware ein/aus, nicht aber die Ints
    if (stat)                                // RX Buffer Interrupt ein
        STI_TSR   = STI_TSR | 0x01;
    else
        STI_TSR   = STI_TSR & ~0x01;
}


////////////////////////////////////////////////////////
// Schaltet die RX Line komplett ab oder an
void rx_full_enable(char stat)
{
    CLI;
    if (stat) {                                 // RX Buffer Interrupt ein
        STI_RSR   = STI_RSR  | 0x01;            // RX ein
        STI_IMRA  = STI_IMRA | RX_INT_MASK;
        STI_PVR   = MODE2_VECTOR | STI_IERA;
        STI_IDR   = STI_IDR  | RX_INT_MASK;     // Ints EIN
    } else {                                    // RX Buffer Interrupt aus
        STI_RSR   = STI_RSR  & ~0x01;           // RX aus
        STI_IMRA  = STI_IMRA & ~RX_INT_MASK;    // Maske aus
        STI_PVR   = MODE2_VECTOR | STI_IERA;
        STI_IDR   = STI_IDR  & ~RX_INT_MASK;    // Enable aus
    }
    STI_IPRA  = STI_IPRA  & ~RX_INT_MASK;       // Pending Ints lschen
    EI;
}

////////////////////////////////////////////////////////
// Schaltet die TX Line ab oder an
void tx_full_enable(char stat)
{
    CLI;
    if (stat) {                               // RX Buffer Interrupt ein
        STI_TSR   = STI_TSR | 0x01;
        STI_IMRA  = STI_IMRA | TX_INT_MASK;   // Int Mask
        STI_PVR   = MODE2_VECTOR | STI_IERA;
        STI_IDR   = STI_IDR  | TX_INT_MASK;   // Int Enable
    } else {                                  // TX Interrupt aus
        STI_TSR   = STI_TSR & ~0x01;
        STI_IMRA  = STI_IMRA & ~TX_INT_MASK;  // Maske aus
        STI_PVR   = MODE2_VECTOR | STI_IERA;
        STI_IDR   = STI_IDR  & ~TX_INT_MASK;  // Enable aus
    }
    STI_IPRA  = STI_IPRA  & ~TX_INT_MASK; // Pending Ints lschen
    EI;
}


////////////////////////////////////////////////////////
// Schaltet Timer A ein/aus
void ta_full_enable(char stat)
{
    CLI;
    if (stat) { // TA Interrupt ein
        // Timer A = endlos Run
        STI_TADR  = 0xb8;                   // Reload Wert Timer A, ca 100 HZ
        STI_TABCR = 0x70;                   // Timer A DIV 200, maximale Verzgerung

        STI_PVR   = MODE2_VECTOR | STI_IERA;
        STI_IDR   = STI_IDR  | 0x20;        // Ints ein
        STI_IMRA  = STI_IMRA | 0x20;        // Mask ein
    } else  {  // TA Interrupt aus
        STI_PVR   = MODE2_VECTOR| STI_IERA;
        STI_IDR   = STI_IDR  & ~0x20;
        STI_IMRA  = STI_IMRA & ~0x20;
    }

    EI;
}


// Interrupts global EIN/AUS
void interrupts(char stat)
{
    if (stat) {
        STI_PVR   = MODE2_VECTOR;           // Vektor auf Mode 2 Tabelle
         __asm
            im 2
            ld a,#CODE_START >> 8 & 0xff
            ld i,a
            ei
        __endasm;
    } else {
        __asm
            di
        __endasm;
    }
}

// Intel Hex Decoder



