#ifndef driver_h
    #define driver_h

// Groesse der Uart Puffer in Bytes
#define BUF_SIZE        128
#define RX_INT_MASK     0x10        // Maske für RX Ints ein
#define TX_INT_MASK     0x04        // Maske für TX Ints ein
#define MODE2_VECTOR    0x40        // zieladresse für Mode 2 Tabelle

// RX Empfangspuffer
struct rx_type{
    int8_t  buf[BUF_SIZE];
    uint8_t rd_ptr;               // Read Pointer, nur durch Hauptprogramm lesen
    uint8_t wr_ptr;               // Write Pointer, nur durch Interrupt schreiben
    uint8_t bytes_to_read;        // Anzahl noch zu lesender Bytes
    uint8_t mask;
    uint8_t f_buf_overflow;       // flag:Overflow: WR Zeiger hat RD überholt: Buffer wertlos
};

// TX Sendepuffer
struct tx_type{
    int8_t  buf[BUF_SIZE];
    uint8_t rd_ptr;               // Read Pointer, nur durch Hauptprogramm lesen
    uint8_t wr_ptr;               // Write Pointer, nur durch Interrupt schreiben
    uint8_t bytes_to_send;        // Anzahl ungesendeter Bytes
    uint8_t mask;
    uint8_t f_idle;               // TX Hardware ist inaktiv
};


/*
    Externe Hardware am Demuxer A5,A6,A7

    Mux 0 : $00:    STI Multi I/O Chip
    Mux 1 : $20:    8255
    Mux 2 : $40:    ADC 0804
    Mux 3 : $60     7 Segment

    I/O Map     0x20..0x2F - STI

    0x00        IDR     Indirect Data Register
    0x01        GPIP    Purpose I/O Data Register
    0x02        IPRB    lnterrupt Pending Register B
    0x03        IPRA    Interrupt Pending Register A .
    0x04        ISRB    Interrupt -Service Register B
    0x05        ISRA    Interrupt in-Service Register A
    0x06        IMRB    lnterrupt Mask Register B
    0x07        IMRA    Interrupt Mask Register A
    0x08        PVR     Pointer,/Vector Register
    0x09        TABCR   Timers A and B Control Register
    0x0A        TBDR    Timer B Data Register
    0x0B        TADR    Timer A Data Register
    0x0C        UCR     USART Control Register
    0x0D        RSR     RX Receiver Status Register
    0x0E        TSR     TX Transmitter Status Register
    0x0F        UDR     USART Data Register

    I/O Map     0x40..0x43 - 8255
    0x20        PORTA
    0x21        PORTB
    0x22        PORTC
    0x23        8255_CNTRL

    I/O Map     0x60 - ADC 0804
    0x40        ADC0804_READ

    I/O Map     0x80 - 7_Segment Anzeige
    0x60        SEG7


*/

// Special Register STI Multi I/O Chip
__sfr __at 0x00 STI_IDR;
__sfr __at 0x01 STI_GPIP;
__sfr __at 0x02 STI_IPRB;
__sfr __at 0x03 STI_IPRA;
__sfr __at 0x04 STI_ISRB;
__sfr __at 0x05 STI_ISRA;
__sfr __at 0x06 STI_IMRB;
__sfr __at 0x07 STI_IMRA;
__sfr __at 0x08 STI_PVR;
__sfr __at 0x09 STI_TABCR;
__sfr __at 0x0a STI_TBDR;
__sfr __at 0x0b STI_TADR;
__sfr __at 0x0c STI_UCR;
__sfr __at 0x0d STI_RSR;
__sfr __at 0x0e STI_TSR;
__sfr __at 0x0f STI_UDR;


// Indirekt über STI_IDR adressbierbare Register des STI
#define     STI_SCR     0       // Sync Character Register
#define     STI_TDDR    1       // Timer D Data Register
#define     STI_TCDR    2       // Timer C Data Register
#define     STI_AER     3       // Active Edge Register
#define     STI_IERB    4       // Interrupt Enable Register B
#define     STI_IERA    5       // Interrupt Enable Register A
#define     STI_DDR     6       // Data Direction Register
#define     STI_TCDCR   7       // Timer C und D Control Register

// PIO 8255 Special I/O
__sfr __at 0x20 PIO8255_PORT_A;
__sfr __at 0x21 PIO8255_PORT_B;
__sfr __at 0x22 PIO8255_PORT_C;
__sfr __at 0x23 PIO8255_CNTRL;

// ADC 0804
__sfr __at 0x40 ADC0804_READ;

// 7b Segment
__sfr __at 0x60 SEG7;

// Prototypen

extern void InitHardware();
extern void clear_buffer(char*,char);
extern char getchar();
extern char keypress();
extern char ihx_decode();
extern void putchar(char);
extern uint8_t adc_read();
extern void Set_GPIO(uint8_t,uint8_t);
extern void SetLed(uint8_t, uint8_t);
extern void SetLedByte(uint8_t);
extern void delay(uint16_t);
extern void stop_cpu();
extern void SetDigit(unsigned char);
extern void interrupts(char);
extern void init_buffer();
extern void rx_full_enable(char);
extern void tx_full_enable(char);
extern void ta_full_enable(char);
extern void rx_fast_enable(char);
extern void tx_fast_enable(char);
extern void delay(uint16_t);
extern uint16_t prng(uint16_t);
extern void sprng(uint16_t);
uint32_t TT800(uint16_t);

// Makros

#define CLI             __asm di __endasm               // Interrupts sperren/frei
#define EI              __asm ei __endasm
#define EI_RETI         __asm ei   \
                              reti \
                        __endasm


#endif
