/////////////////////////////////////////////////////////
/*
   Minimum Z80 System mit Mostek 3801 Multi I/O Chip
   (C)Christian Julius 10.2014
   christian.julius@gmx.de

   Interrupt Handler Routine Mode 2
   siehe crt0.s fr Vektoren

*/
/////////////////////////////////////////////////////////

#include <stdio.h>
#include <stdint.h>
#include <stdbool.h>
#include "main.h"
#include "driver.h"

extern struct system_type sys;             // System Infos

// RX und TX Buffer....
extern struct tx_type tx;
extern struct rx_type rx;

/*  Interrupt Zusatz:
  __interrupt erzeugt ein reti am Ende, Register werden gesichert
  __naked verhindert Erzeugung von Code zur Registersicherung auf dem Stack
   und erzeugt kein RETI
*/

////////////////////////////////////////////////////////
// NMI Handler siehe crt0.s Definition
// __critical __interrupt erzeugt ein retn am Ende statt reti

void nmi_vint(void) __critical __interrupt
{
     __asm
         di
         ld a,#0               // Lsche User Programm Kennung
         ld (RAM_START),a
         jp ROM_START          // Software Reset des Z80
    __endasm ;

}

/////////////////////////////////////////////////////////
// Handler der Mode 2 Interrupts der STI
//
// ISR: Uart RX puffer (1 Byte) ist voll, es sind Zeichen da!
// getchar() muss diese abholen und Zeiger verstellen
// f_buf_overflow muss vom Hauptprogramm zurck gesetzt werden

void int_sti_receive_buffer_full(void)  __interrupt  // geprft = ok
{
    uint8_t zeichen;

    zeichen = STI_UDR;
    // Noch Platz im Puffer?
    if (rx.bytes_to_read < BUF_SIZE)
    {
        rx.buf[rx.wr_ptr] = zeichen;                // Zeichen in Ringpuffer schreiben
        rx.wr_ptr = (rx.wr_ptr +1) & rx.mask;       // Rotieren
        rx.f_buf_overflow = false;
        rx.bytes_to_read++;
    } else
        rx.f_buf_overflow = true;                   // Buffer Overflow

    EI;
}

////////////////////////////////////////////////
// ISR: Uart TX Sendepuffer ist leer
// putchar()fllt ihn wieder
//
// Vernderte globale Variablen:
// tx.bytes_to_send, tx.rd_ptr

void int_sti_transmit_buffer_empty(void)  __interrupt
{
    // Sind noch Zeichen im TX Buffer?
    if (tx.bytes_to_send > 0)  {                // ja.....
        STI_UDR   = tx.buf[tx.rd_ptr];          // Byte einschreiben
        tx.rd_ptr = (tx.rd_ptr +1) & tx.mask;   // Read Pointer rollen
        LedOut(tx.bytes_to_send);
        tx.f_idle = false;
        tx.bytes_to_send--;
    }
    else {
        // nein, letztes Zeichen ging grad raus
        tx.f_idle = true;
    }

    EI;

}

////////////////////////////////////////////////////////////////
// Interrupt des Timers A (Hardware Timeout)
/*
    Funktionsweise

    Mit der Variablen TA_RESET wird eine Zeit in 1/10s eingestellt.
    TA_Enable = 1 => Watchdog ist aktiv
    Wird die globale Variable TA_Timeout nicht stndig auf 0 zurck gesetzt durch
    die Anwendung luft der timer A ber und lst einen Software Reset aus

*/


static void tadelay(uint16_t cnt)
{
    volatile uint16_t k;
    for (k=0;k<cnt;k++);
}

extern uint8_t AD_Value;

void int_sti_timer_a(void) __interrupt
{
    static uint16_t TA_Reset = 200;   // 1/10 Sekunden
    static uint16_t counter = 0;
    char i;
    static char toggle1 = 0;
    static char toggle2 = 0;

    /*
    // AD Wandler auslesen
    if (toggle2) {
        ADC0804_READ = 1;           // Wandlung triggern
        toggle2 = 1;
    }
    else {
        AD_Value = ADC0804_READ;    // Wandlung auslesen
        toggle2 = 0;
    }
*/
    // Watchdog bedienen
    if (sys.f_TA_Enable == true)
    {
        if (++counter > 5) {
            counter = 0;
            if (toggle1) {
                Set_GPIO(7,HIGH);
                toggle1 = 0;
            }
            else {
                Set_GPIO(7,LOW);
                toggle1 = 1;
            }

            // berlauf?
            if (sys.TA_Timeout++ >= TA_Reset) {
             CLI;
              // LEDs blinken lassen fr Reset
             for (i=0;i<10;i++) {
                LedOut(0);
                tadelay(2000);
                LedOut(0xff);
                tadelay(2000);
             }

              // Soft reset
                __asm
                    di
                    jp 0x0000
                __endasm;
            }
        }
    }

    EI;
}


////////////////////////////////////////
// ------- Unbenutzte Ints -------------
////////////////////////////////////////


// AD Wandler Interrupt ber STI
void int_sti_gpi_0(void) __interrupt
{
   EI;
}

// Interrupt des Timers B
void int_sti_timer_b(void)  __interrupt __naked
{
   EI_RETI;
}

// Interrupt des Timers C
void int_sti_timer_c(void)  __interrupt __naked
{
    EI_RETI;
}

void int_sti_timer_d(void)  __interrupt __naked
{
    EI_RETI;
}

void int_sti_gpi_1(void)  __interrupt __naked
{

    EI_RETI;
}
void int_sti_gpi_2(void)  __interrupt __naked
{
    EI_RETI;
}
void int_sti_gpi_3(void)  __interrupt __naked
{
    EI_RETI;
}

void int_sti_gpi_4(void)  __interrupt __naked
{
   EI_RETI;
}

void int_sti_gpi_5(void)  __interrupt __naked
{
    EI_RETI;
}

void int_sti_gpi_6(void)  __interrupt __naked
{
    EI_RETI;
}

void int_sti_gpi_7(void)  __interrupt __naked
{
    EI_RETI;
}

void int_sti_transmit_error(void)  __interrupt __naked
{
    EI_RETI;
}

void int_sti_receive_error(void)  __interrupt __naked
{
    EI_RETI;
}






