#include <stdio.h>
#include <stdlib.h>
#include <stdint.h>
#include <stdbool.h>
#include <string.h>
#include "main.h"
#include "driver.h"
#include "sd.h"
#include "mcurses.h"

/////////////////////////////
// Externals

extern struct system_type sys;

//#define DELAY_ON

// Eigene optionale Verzögerung
static void delay8(uint8_t cnt) {
    volatile uint8_t k;
    for (k=0;k<cnt;k++) {
        __asm
        nop
        __endasm;
    }

}

// Einstellungen für SPI
void init_spi()
{
    CLI;
    // Interrupts ausschalten (Figure 8, MK3801 Manual, ind. IERB Register)
    STI_PVR = STI_IERB;
    STI_IDR = STI_IDR & ~( (1<<1) | (1<<2) | (1<<3) | (1<<6) | (1<<7) );

    // Data Direction einstellen (Figure 14, MK3801 Manual)
    STI_PVR = STI_DDR;
    STI_IDR = STI_IDR | SDO | SCLK | SS;    // 1 = SDO, SCLK, SS = Outputs
    STI_IDR = STI_IDR & ~(SDI | WAIT);      // 0 = SDI, WAIT = Inputs

    // Vektor zurückschreiben
    STI_PVR   = MODE2_VECTOR;
    EI;

    // Voreinstellungen
    STI_GPIP = STI_GPIP | SCLK;             // Clock = HIGH
    STI_GPIP = STI_GPIP | SS;               // SS = HIGH
    STI_GPIP = STI_GPIP & ~SDO;             // SDO = 0
}

// Fragt die WAIT Leitung ab
// 1 = Master muss warten
// 0 = Master kann weitersenden
char spi_get_wait()
{
    return (STI_GPIP & WAIT);
}

/////////////////////////////////////////////////////
// Setzt die Chipselect Leitung des SD Karten Slave
void sd_select(uint8_t stat)
{
    if (stat) {
        STI_GPIP = STI_GPIP & ~SS;      // Slave Select LOW
        #ifdef DELAY_ON
            delay8(PAUSE);
        #endif
    }
    else {
        #ifdef DELAY_ON
            delay8(PAUSE);
        #endif
        STI_GPIP = STI_GPIP | SS;      // Slave Select HIGH
    }
}

/*
 * Schema einer Bit Übertragung mit
 * CPOL=1 (high=inaktiv), CPHA = 1 (lo->hi)
 * MSB first
 *
 * 1. SCLK = HIGH   => Clock inaktiv
 * 2. CS = LOW      => Slave aufwecken
 * 3. delay
 * 4. SCLK = LOW            Slave legt Bit an
 * 5. SDO = Bit             Master legt Bit an
 * 5. SCLK = HIGH           Master und Slave Bit ist jetzt gültig!
 * 6. Databit = SDI         Master sampelt Bit vom Slave, Slave das vom Master
 * 6. 8 x weiter bei 4
  */

///////////////////////////////////////////////
// Ebene: 0
// Ein Byte senden und empfangen
//
uint8_t spi_transmit(uint8_t data)
{

    uint8_t inbyte;
    uint8_t i;

    STI_GPIP = STI_GPIP | SCLK;     // Clock = HIGH
    inbyte   = 0;

    CLI;
    for (i=0;i<8;i++)
    {
       // ----- Clock LOW  (Slave sampled) ------
        #ifdef DELAY_ON
         delay8(PAUSE);
        #endif
         STI_GPIP = STI_GPIP & ~SCLK;

        // // ----- Bit out SDO  ------
        if (data & 0x80)
           STI_GPIP = STI_GPIP | SDO;      // "1" out
        else
           STI_GPIP = STI_GPIP & ~SDO;     // "0" out

         // // ----- Clock HIGH  ------
        #ifdef DELAY_ON
         delay8(PAUSE);
        #endif
         STI_GPIP = STI_GPIP | SCLK;

        // // ----- Sample Bit In SDI  ------
         if (STI_GPIP & SDI)        // Eine 1 Input?
              inbyte |=  0x01;      // 0 egal, Inbyte = 0

        // Nächste Position, MSB zuletzt
        data   <<= 1;
        inbyte <<= 1;
    }
    STI_GPIP = STI_GPIP & ~SDO;        // "0" out
    STI_GPIP = STI_GPIP | SCLK;        // SCLK = 1

    EI;

    return (inbyte);
}

/////////////////////////////////////////////////////////////
// Ebene 1
// SD Funktionen (Protokoll)
/////////////////////////////////////////////////////////////

#define TYPE_PROG    0
#define TYPE_BIN     1
#define TYPE_ASC     2

#define CMD_HEAD 'H'        // Token für File schreiben
#define CMD_WR   'W'        // Token für File schreiben
#define CMD_RD   'R'        // Token für File lesen
#define CMD_D    'D'        // Token für File löschen
#define DUMMY     0xff
#define RES_ACK  'A'        // Response für "OK"
#define RES_WAIT 'T'        // Response für "Warte!"
#define RES_ERR  'E'        // Response für "Warte!"

/*
    Der Slave (Arduino) nimmt sobald die SS Leitung LOW geht einen
    fixen 64 Bytes Header Datensatz entgegen. Danach zieht er direkt
    WAIT = HIGH.

    Aus diesem Header entnimmt er alle Information was mit den
    nachfolgenden Daten zu machen ist

    Basisfunktionen für Menu:

    sd_save_prog(Name, Beschreibung)        Sichere  den RAM Dump auf Karte
    sd_catalog                              Liste Files der Karte auf
    sd_load(Filename, Adresse)              Lade File <name> von Karte in den Speicher
    sd_delete(Filename)                     Lösche File von SD Karte

    Funktionen zur Verwendung in Programmen

    sd_open_file(Name,Mode R/W)             Öffnet ein File zum Schreiben oder Lesen
    sd_write_file(Name,Zeiger auf Daten, Datenlänge);   Schreibt Daten in ein File ein
    sd_print_file(Name, String)             Schreibt einen String in ein File
    sd_seek_file(Name, Zeiger)              Setzt den Zeiger auf Byte Nr. n
    sd_file_info(Name)                      Holt den Info Header eines Files
    sd_read_file(Name,&Byte)                Liest ein Byte aus einem File. Rück: EOF


*/

// File-header, Laenge: 64 Bytes
struct file_header{
      uint8_t   cmd;              // 1) Kommando für Aktion, WR RD DL usw
      uint8_t   type;             // 1) 0 = Programm, 1 = Datensatz, 2 = ASCII
      uint16_t  start_adress;     // 2) Programm: Ursprüngliche Startadresse im Speicher (0 = keine)
      uint16_t  end_adress;       // 2) Programm: Ursprüngliche Startadresse im Speicher (0 = keine)
      uint16_t  pointer;          // Zeiger auf Fileposition
      uint32_t  size;             // Groesse des Files  (nur für Info lesen)
      uint32_t  len;              // 4) Laenge des Datensatzes in Bytes
      uint8_t   name[12];         // 8) Dateiname
      uint8_t   descr[32];        // 32) Beschreibung des Files
      uint16_t  crc;              // Prüfsumme (einfache 16 Bit Addition)
      uint8_t   dummy[2];         // Füllbytes auf 64
};

// Overlay des RAM Speichers
__at (RAM_START) uint8_t ramloc[RAM_SIZE];          // Bereich des RAM

/* /////////////////////////////////////////////////////////

   Ebene 1: sd_save_prog

   char* adr
   adr = &ram[0];
   sd_save_prog(1,"test.dat","Ein Testfile",adr,(uint16_t)adr,0x4000);

   Schreibt einen Datensatz auf die SD Karte

   Eingabe:
           file       =  filenummer 1..255
           *filename  =  Zeiger auf Filename
           *filedescr =  Beschreibung
           *data      =  Zeiger auf Daten
           adr        =  Startadresse der Daten
           len        =  Länge der Daten in Bytes

    Rückgabe: 0  =  Fehler, 1 = ok, 2 = File vorhanden

/////////////////////////////////////////////////////////// */

//#define SD_SAVE_PROG_DEBUG

char sd_save_prog(char* filename,  char* filedescr)
{
    uint16_t i,crc;
    uint8_t resp;                   // Antwort des Slaves
    char* ptr;
    struct file_header* sptr;

    // Hole Platz vom Heap für Header
    #ifdef SD_SAVE_PROG_DEBUG
    addstr("\r\r--- START sd_save_prog---");
    addstr("\rAlloziere Heap sptr");
    #endif

    sptr = (struct file_header*) malloc(sizeof(struct file_header));
    if (sptr == NULL)
        return (0);

    #ifdef SD_SAVE_PROG_DEBUG
    xprintf("\r   sizeof(sprtr) = %d",sizeof(struct file_header));
    xprintf("\r   &sprtr = %x",(uint16_t)sptr);
    #endif

    #ifdef SD_SAVE_PROG_DEBUG
        addstr("\rBaue File Header");
    #endif

    // Berechne Prüfsumme vorher
    crc = 0;
    for (i=0;i < sys.code_length; i++)
        crc = crc + ramloc[i];

    #ifdef SD_SAVE_PROG_DEBUG
        addstr("\rPruefsumme berechnet");
    #endif

    // Header Infos zusammenbauen
    sptr->cmd = CMD_WR;
    sptr->type = TYPE_PROG;                         // Datentyp Programm
    sptr->crc  = crc;                               // Prüfsumme

    sptr->start_adress = sys.code_start;                // Startadresse
    sptr->end_adress   = sys.code_end;                  // Endadresse
    sptr->len          = sys.code_length;               // Datenlänge
    strcpy(sptr->name,filename);                    // Dateiname
    strcpy(sptr->descr,filedescr);                  // Dateibeschreibung

    #ifdef SD_SAVE_PROG_DEBUG
    xprintf("\r   sptr->name              = %s",sptr->name);
    xprintf("\r   sptr->descr             = %s",sptr->descr);
    xprintf("\r   sptr->start_adress      = 0x%x",sptr->start_adress);
    xprintf("\r   sptr->end_adress        = 0x%x",sptr->end_adress);
    xprintf("\r   sptr->len               = %u Bytes",sptr->len);
    xprintf("\r   sptr->crc               = %x",sptr->crc);
    #endif

    sd_select(true);
    delay8(100);

    // Header in einem Rutsch senden, Slave weiss dass er 64 Bytes ist
    #ifdef SD_SAVE_PROG_DEBUG
        addstr("\rSende Fileheader\r");
    #endif

    // Slave muss antworten
    //    do {
    //        resp = spi_transmit(CMD_HEAD);
    //    } while (resp!=RES_ACK);

    ptr = (char*) sptr;
    for (i=0;i<sizeof(struct file_header);i++) {
        while(spi_get_wait());              // WAIT vom Slave?
        resp = spi_transmit(*(ptr++));
    }

    // Ganzen Datensatz senden
    #ifdef SD_SAVE_PROG_DEBUG
        xprintf("\rSende %u Bytes Rohdaten\r",sptr->len);
    #endif
    for (i=0;i < sptr->len;i++)
    {
        while(spi_get_wait());              // WAIT vom Slave?
        LedOut(ramloc[i]);
        resp = spi_transmit(ramloc[i]);
    }

    //do {
    //      resp = spi_transmit(CMD_HEAD);
    //} while (resp!=RES_ACK);

    delay8(50);
    sd_select(false);

    #ifdef SD_SAVE_PROG_DEBUG
        addstr("\r---END sd_save_prog ----\r");
    #endif

    free(sptr);     // Header verwerfen

    return (1);
}



//////////////////////////////////////////////////////////////
//  Öffne ein File
//
//  Params: Nummer des Files, Mode: Write/Read






